#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <pcl/io/vtk_lib_io.h>

#include "types.h"
#include "utils.h"
#include "geom_utils.h"
#include "io_utils.h"
#include "vis_wrapper.h"

#include "eigen_extensions.h"

using namespace std;

// Loads the entire bag, a result of running segment_scene.cpp on many scenes
// and allows for annotating segments as true or false
// Outputs a big NxD matrix fof measures for entire bag (as columns), 
// together with an Nx1 vector of labels (+1 or -1). 

// REQUIRES: segment_scene.cpp was run on some scenes (i.e. /measures and /segments is non-empty)
// REQUIRES: cooccurrence.cpp was run to fill in co-occurrence measure column

// PRODUCES: the labels "segmentLabels.eig.txt" vector via user interaction
int main(int argc, char** argv) {
  srand(1);

  int MAXSCENEID= 70; // highest id of any scene
  
  int loadLabels= 0;
  if(argc>1) loadLabels = atoi(argv[1]);
  
  // Go over all scenes, load all segments and measures
  vector<Cloud_t::Ptr> allclouds;
  vector<NormalCloud_t::Ptr> allnormals;
  vector<int> sceneids;
  int NS=0;
  int N=0;
  for(int i=0;i<MAXSCENEID;i++) {
    string plyfile= getScenePath(i);
    if(!file_exists(plyfile)) continue; // skip non-existing scene
    int g= 0;
    
    // Load in all segments as .ply files
    string segfile= getSegmentPath(i, g);
    while(file_exists(segfile)) {
      
      Mesh_t::Ptr mesh (new Mesh_t);
      Cloud_t::Ptr cloud (new Cloud_t);
      NormalCloud_t::Ptr normals(new NormalCloud_t);
      pcl::io::loadPolygonFile(segfile, *mesh);
      mesh_to_points_normals(mesh, cloud, normals);
      allclouds.push_back(cloud);
      allnormals.push_back(normals);
      sceneids.push_back(i);
      
      g++;
      segfile= getSegmentPath(i, g);
    }
    
    printf("Loaded scene %d/%d with %d segments...\n", i, MAXSCENEID, g);
    NS += g;
    N++;
  }
  printf("number of segments: %d from %d scenes.\n", NS, N);
  
  // load measures from file (created by cooccurrence)
  string measuresPath = ROOT_PATH + string("segmentMeasures.eig.txt");
  Eigen::MatrixXf measures;
  eigen_extensions::loadASCII(measuresPath, &measures);
  assert(measures.rows() == NS);
  Eigen::VectorXf objectness= measures.rowwise().sum() / measures.cols(); // objectness is just the average
  
  // sort according to objectness (average across measures)
  vector<fipair> objv(NS);
  for(int i=0;i<NS;i++) objv[i]= fipair(objectness(i), i);
  sort(objv.begin(), objv.end(), fiComparatorDescend);
  vector<Cloud_t::Ptr> cs;
  vector<NormalCloud_t::Ptr> ns;
  for(int i=0;i<NS;i++) { int ix= objv[i].second; cs.push_back(allclouds[ix]); ns.push_back(allnormals[ix]); }
  
  // start the visualizer
  VisWrapper v;
  v.vis_.setBackgroundColor(0.5, 0.5, 0.5);
  v.vis_.setCameraPose(0, 0, 0, 
                       0.3, 0.3, 0.3, 
                       0, -1, 0);
                       
  // show the cloud
  showCloudsGridOriented(v, cs, ns);
  //while(v.waitKey() != 'q') continue;
  
  // initialize labels
  string labelsPath = ROOT_PATH + string("segmentLabels.eig.txt"); // these are the paths for output (and input, in case of labels, potentially)
  Eigen::VectorXf labels(NS);
  labels.setConstant(-1);
  if(loadLabels==1) {
    if(file_exists(labelsPath)) {
      eigen_extensions::loadASCII(labelsPath, &labels);
      printf("loaded labels from file.\n");
    } else {
      printf("attempt was made to load labels but the labels file doesn't exist.\n");
    }
  }
  
  // show the labeling interface
  int s=0;
  Eigen::Quaternionf rot; rot.setIdentity();
  bool redraw= true;
  float space= 0.5;
  int gsize= (int) sqrt(NS);
  while(true) {
  
    // draw selection boxes
    if(redraw) {
      v.vis_.removeAllShapes();
      for(int i=0;i<NS;i++) {
        int j= objv[i].second;
        if(labels(j)==1) {
          Eigen::Vector3f tvec;
          tvec << (i/gsize) * space, (i%gsize) * space, 0;
          v.vis_.addCube(tvec, rot, space, space, space, "box_" + to_str(i));
        }
      }
      redraw=false;
    }
    
    // draw the selection box
    v.vis_.removeShape("selectionbox");
    Eigen::Vector3f tvec;
    tvec << (s/gsize) * space, (s%gsize) * space, 0;
    v.vis_.addCube(tvec, rot, space/2, space/2, space/2, "selectionbox");
    
    // print some info about selected segment
    int ir= objv[s].second;
    cout << "segment from scene " << sceneids[ir] << ", measures: " << measures.row(ir) << endl;
    
    // handle user itneraction
    char c= v.waitKey();
    int olds= s;
    if(c=='q') break;
    if(c=='w') { s-= 1; }
    if(c=='s') { s+= 1; }
    if(c=='a') { s-= gsize; }
    if(c=='d') { s+= gsize; }
    if(s<0 || s>=NS) s=olds; // reset it, invalid movement along grid
    
    if(c=='x') { // toggle label on this segment

      if(labels(ir)==1) {
        labels(ir)=-1;
        // we are toggling the box off, remove it
        v.vis_.removeShape("box_" + to_str(s));
      } else {
        labels(ir)=1;
        Eigen::Vector3f tvec;
        tvec << (s/gsize) * space, (s%gsize) * space, 0;
        v.vis_.addCube(tvec, rot, space, space, space, "box_" + to_str(s));
      }
    }
  }
  
  // save the labels and the normalized measures
  eigen_extensions::saveASCII(labels, labelsPath);
  printf("Saved labels and measures!!\n");
  printf("done.\n");
  return(0); 
}
