#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <pcl/io/vtk_lib_io.h>

#include "types.h"
#include "utils.h"
#include "geom_utils.h"
#include "io_utils.h"
#include "vis_wrapper.h"

#include "eigen_extensions.h"

using namespace std;

// REQUIRES: segment_scene.cpp was run on some scenes (i.e. /measures and /segments is non-empty)
// PRODUCES: segmentMeasures.eig.txt (contains all measures, including 
//           co-occurrence for all segments concatenated in a single big matrix)
int main(int argc, char** argv) {
  srand(1);

  int MAXSCENEID= 70; // highest id of any scene
  float MULTPLY = 0.25;
  int NNEED= 10; // number of nearest neighbors to consider when computing co-occurence score
  int DISPLAY= 0;
  if(argc>1) DISPLAY= atoi(argv[1]);
  
  // Go over all scenes, load all segments and measures
  Eigen::MatrixXf measures;
  vector<Cloud_t::Ptr> allclouds;
  vector<NormalCloud_t::Ptr> allnormals;
  vector<int> sceneids;
  int NS=0; // number of segments
  int N=0; // number of scenes
  for(int i=0;i<MAXSCENEID;i++) {
    string plyfile= getScenePath(i);
    if(!file_exists(plyfile)) continue; // skip non-existing scene
    
    // Load in all segments as .ply files
    int g= 0;
    string segfile= getSegmentPath(i, g);
    while(file_exists(segfile)) {
      
      Mesh_t::Ptr mesh (new Mesh_t);
      Cloud_t::Ptr cloud (new Cloud_t);
      NormalCloud_t::Ptr normals(new NormalCloud_t);
      pcl::io::loadPolygonFile(segfile, *mesh);
      mesh_to_points_normals(mesh, cloud, normals);
      allclouds.push_back(cloud);
      allnormals.push_back(normals);
      sceneids.push_back(i);
      
      g++;
      segfile= getSegmentPath(i, g);
    }
    
    // Load in the measures for the .ply files and stack them as rows into one big matrix
    Eigen::MatrixXf measuremat;
    string measurepath= getMeasurePath(i);
    eigen_extensions::loadASCII(measurepath, &measuremat);
    measures.conservativeResize(measures.rows() + measuremat.rows(), measuremat.cols());
    measures.block(measures.rows() - measuremat.rows(), 0, measuremat.rows(), measuremat.cols()) = measuremat; // insert it on bottom
    
    printf("Loaded scene %d/%d with %d segments...\n", i, MAXSCENEID, g);
    NS += g;
    N++;
  }
  printf("number of segments: %d from %d scenes. Size of measures matrix: %d x %d\n", NS, N, (int)measures.rows(), (int)measures.cols());
  assert(NS==((int)measures.rows()));
  
  // normalize the measures for all segments
  printf("normalizing measures...\n");
  Eigen::MatrixXf mm(measures);
  Eigen::VectorXf mmean= mm.colwise().sum() / mm.rows();
  mm.rowwise() -= mmean.transpose();
  Eigen::MatrixXf mstd= (mm.colwise().squaredNorm() / mm.rows()).cwiseSqrt();
  
  // Normalize the data
  Eigen::MatrixXf measuresNormalized(measures);
  measuresNormalized.rowwise() -= mmean.transpose();
  measuresNormalized = measuresNormalized.cwiseQuotient(mstd.replicate(measuresNormalized.rows(), 1));
  //Eigen::VectorXf objectness= measuresNormalized.rowwise().sum() / measuresNormalized.cols(); // objectness is just the average
  
  // get spatial extents of all segments
  printf("finding spatial extents of all segments...\n");
  Eigen::MatrixXf V(NS, 3);
  for(int i=0;i<NS;i++) {
    Eigen::MatrixXf ptsmap = allclouds[i]->getMatrixXfMap();
    Eigen::VectorXf mins= ptsmap.rowwise().minCoeff(); 
    Eigen::VectorXf maxes= ptsmap.rowwise().maxCoeff(); 
    Eigen::VectorXf ranges= maxes - mins;
    V.row(i)= ranges.head(3); // save the vx, vy, vz
  }
  
  // start the visualizer
  VisWrapper v;
  v.vis_.setBackgroundColor(0.5, 0.5, 0.5);
  v.vis_.setCameraPose(0, 0, 0, 
                       0.3, 0.3, 0.3, 
                       0, -1, 0);
  
  printf("computing co-occurrence score\n");
  Eigen::VectorXf score(NS);
  for(int i=0;i<NS;i++) {
    
    float mult= MULTPLY;
    Eigen::VectorXf vthr= V.row(i) * mult;
    vector<fipair> x;
    while(x.size() < NNEED) {
    
      x.clear();
      for(int k=0;k<NS;k++) {
        if(sceneids[i] == sceneids[k]) continue;
        if(((V.row(i)-V.row(k)).transpose().array().abs() > vthr.array()).any()) continue;
        float dd= (measuresNormalized.row(i).tail(4) - measuresNormalized.row(k).tail(4)).array().square().sum();
        x.push_back(fipair(dd, k));
      }
      mult += 0.02;
      vthr= V.row(i) * mult;
    }
    sort(x.begin(), x.end(), fiComparatorAscend);
    float coscore= 0.0;
    int k=0;
    int thisscene= sceneids[i];
    int numn=0;
    vector<Cloud_t::Ptr> toshow;
    vector<NormalCloud_t::Ptr> toshown;
    toshow.push_back(allclouds[i]);
    toshown.push_back(allnormals[i]);
    while(numn<NNEED && numn<x.size()) { 
      coscore += x[k].first; 
      if(DISPLAY==1) { printf("adding %f to coscore!\n", x[k].first); }
      toshow.push_back(allclouds[x[k].second]);
      toshown.push_back(allnormals[x[k].second]);
      
      k++;
      numn++;
    }
    coscore /= numn;
    printf(" SUM=%f based on %d measurements and extent %f for %d/%d\n", coscore, numn, mult, i, NS);
    if(DISPLAY==1) {
      showCloudsGridOriented(v, toshow, toshown);
      while(v.waitKey()!='q') continue;
    }
    score(i)= -coscore;
  }
  
  // visualize the final ordering
  if(DISPLAY==2) {
    vector<fipair> objv(NS);
    for(int i=0;i<NS;i++) objv[i]= fipair(score(i), i);
    sort(objv.begin(), objv.end(), fiComparatorDescend);
    vector<Cloud_t::Ptr> cs;
    vector<NormalCloud_t::Ptr> ns;
    for(int i=0;i<NS;i++) { int ix= objv[i].second; cs.push_back(allclouds[ix]); ns.push_back(allnormals[ix]); }
    showCloudsGridOriented(v, cs, ns);
    while(v.waitKey() != 'q') continue;
  }
  
  // save the measures that include co-occurrence as last column in a file "segmentMeasures.etc.txt"
  string measuresPath = ROOT_PATH + string("segmentMeasures.eig.txt");
  measures.conservativeResize(measures.rows(), measures.cols()+1);
  measures.col(measures.cols()-1)= score;
  
  // normalize before saving
  Eigen::VectorXf mfMean= measures.colwise().sum() / measures.rows();
  measures.rowwise() -= mfMean.transpose();
  Eigen::MatrixXf mfStd= (measures.colwise().squaredNorm() / measures.rows()).cwiseSqrt();
  measures = measures.cwiseQuotient(mfStd.replicate(measures.rows(), 1));
    
  eigen_extensions::saveASCII(measures, measuresPath);
  cout << "saving final measures to " << measuresPath << endl;

  return(0); 
}
