#ifndef __OBJ_GEOM_UTILS_H__
#define __OBJ_GEOM_UTILS_H__

#include "types.h"
#include <vector>
#include "vis_wrapper.h"
#include <pcl/common/pca.h>

using namespace std;
using namespace Eigen;

void mesh_to_points_normals( const Mesh_t::ConstPtr &mesh, Cloud_t::Ptr &cloud, NormalCloud_t::Ptr &normals) {
    
  // unfold mesh, which is stored as ros message
  pcl::fromROSMsg( mesh->cloud, *cloud );
  vector<int> counts(cloud->points.size(), 0);
  normals->points.resize(cloud->points.size());
  
  for(size_t i=0;i<mesh->polygons.size();i++) {
    
    const pcl::Vertices& vv= mesh->polygons[i];
    
    // Get the 3 points
    int i1= vv.vertices[0];
    int i2= vv.vertices[1];
    int i3= vv.vertices[2];
    Point_t& p1= cloud->points[i1];
    Point_t& p2= cloud->points[i2];
    Point_t& p3= cloud->points[i3];
    
    // Convert to eigen points
    Vector3d pe1(p1.x,p1.y,p1.z);
    Vector3d pe2(p2.x,p2.y,p2.z);
    Vector3d pe3(p3.x,p3.y,p3.z);
    
    // Find normal
    Vector3d normal= (pe2-pe1).cross(pe3-pe1);
    normal= normal/normal.norm();
    Normal_t pnormal( normal[0], normal[1], normal[2] );
    
    // smoothly blend with the old normal estimate at this point. Basically each face votes for the normal
    // of the verteces around it.
    float v;
    Normal_t a;
    a = normals->points[i1];
    v= 1.0/(counts[i1]+1.0);
    normals->points[i1]= Normal_t( v*pnormal.normal_x + (1.0-v)*a.normal_x, v*pnormal.normal_y + (1.0-v)*a.normal_y, v*pnormal.normal_z + (1.0-v)*a.normal_z);
    
    a = normals->points[i2];
    v= 1.0/(counts[i2]+1.0);
    normals->points[i2]= Normal_t( v*pnormal.normal_x + (1.0-v)*a.normal_x, v*pnormal.normal_y + (1.0-v)*a.normal_y, v*pnormal.normal_z + (1.0-v)*a.normal_z); 
    
    a = normals->points[i3];
    v= 1.0/(counts[i3]+1.0);
    normals->points[i3]= Normal_t( v*pnormal.normal_x + (1.0-v)*a.normal_x, v*pnormal.normal_y + (1.0-v)*a.normal_y, v*pnormal.normal_z + (1.0-v)*a.normal_z);  
    
    counts[i1]++; counts[i2]++; counts[i3]++; 
  }
}

// return all edges, WITH duplicates 
void getEdges(Mesh_t::Ptr mesh, vector<pair<int,int> > &e) {
  for(int i=0; i < mesh->polygons.size(); i++) { 
      const pcl::Vertices& v= mesh->polygons[i];
      int v1 = v.vertices[0];
      int v2 = v.vertices[1];
      int v3 = v.vertices[2];
      e.push_back(pair<int, int>(v1,v2));
      e.push_back(pair<int, int>(v1,v3));
      e.push_back(pair<int, int>(v3,v2));
  }
}

struct rgb{
  int r;
  int g;
  int b;
};

void showCloudsGrid(VisWrapper &v, vector<Cloud_t::Ptr> &clouds) {
    float space= 0.5;
    int gsize= (int) sqrt(clouds.size());
    Eigen::Affine3f trans;
    v.holdOff();
    for(int i=0;i<clouds.size();i++) {
      Cloud_t::Ptr cdisp(new Cloud_t (*(clouds[i])) );
      trans = Eigen::Translation3f((i/gsize) * space, (i%gsize) * space, 0);
      pcl::transformPointCloud(*(cdisp), *(cdisp), trans);
      v.showCloud(cdisp, 0.02, false);
      v.holdOn();
    }
    v.vis_.spinOnce();
}

void showCloudsGridOriented(VisWrapper &v, vector<Cloud_t::Ptr> &clouds, vector<NormalCloud_t::Ptr> &normals) {
    float space= 0.5;
    int gsize= (int) sqrt(clouds.size());
    Eigen::Affine3f trans;
    v.holdOff();
    for(int i=0;i<clouds.size();i++) {
      Cloud_t::Ptr cdisp(new Cloud_t (*(clouds[i])) );
      
      // flip cloud to face viewer with as many normals as possible along z direction
      float zmass=0;
      for(int k=0;k<cdisp->points.size();k++) {
        Normal_t &n= normals[i]->points[k];
        zmass += n.normal_z;
      }
      if(zmass > 0) {
        // flip it
        for(int k=0;k<cdisp->points.size();k++) {
          cdisp->points[k].z= -cdisp->points[k].z;
        }
      }
      
      trans = Eigen::Translation3f((i/gsize) * space, (i%gsize) * space, 0);
      pcl::transformPointCloud(*(cdisp), *(cdisp), trans);
      v.showCloud(cdisp, 0.02, false);
      v.holdOn();
    }
    v.vis_.spinOnce();
}

// takes cloud, normals and projects them to eigenbasis
// returns result in projcloud, projnormals, 
// and the lambda1normalized and lambda2normalized are second and third eigenvalues of the scatter matrix 
// divided by the first. (so lambda2 < lambda1)
void projectToEigenbasis(
        Cloud_t::Ptr cloud, NormalCloud_t::Ptr normals,
        Cloud_t::Ptr &projcloud, NormalCloud_t::Ptr &projnormals, float &lambda1normalized, float &lambda2normalized) {
  
    pcl::PCA<Point_t> pca;
    pca.setInputCloud(cloud);
    Eigen::Matrix3f eigvecs= pca.getEigenVectors();
    float det= pcl::determinant3x3Matrix(eigvecs);
    if(det < 0) eigvecs = -eigvecs; // Guarantee a rotation. sigh PCL!
    
    Eigen::Vector4f segmean;
    segmean = Eigen::Vector4f::Zero();
    pcl::compute3DCentroid(*cloud, segmean);
    int nn= cloud->points.size();
    Cloud_t::Ptr proj(new Cloud_t);
    NormalCloud_t::Ptr nproj(new NormalCloud_t);
    proj->points.resize(nn);
    nproj->points.resize(nn);
    for(int j=0;j<nn;j++) {
      
      // project points (after subbing the mean)
      Eigen::Vector3f di= cloud->points[j].getVector3fMap() - segmean.head<3>();
      proj->points[j].getVector3fMap() = eigvecs.transpose() * di;
      
      // project the normals
      nproj->points[j].getNormalVector3fMap() = eigvecs.transpose() * normals->points[j].getNormalVector3fMap();
    }
    proj->width= 1;
    proj->height= nn;
    nproj->width= 1;
    nproj->height= nn;
    
    // we have to do funny business because pca doesn't preserve color of points... 
    // also we don't just copy over rgb in case we dont want to use color in future some point
    Cloud_t::Ptr newcloud(new Cloud_t(*cloud));
    for(int j=0;j<proj->points.size();j++) { 
      Point_t &p1= newcloud->points[j];
      Point_t &p2= proj->points[j];
      p1.x= p2.x; p1.y= p2.y; p1.z= p2.z;
    }
    newcloud->width= 1;
    newcloud->height= proj->points.size();
    
    Eigen::Vector3f eigvals= pca.getEigenValues();
    eigvals /= eigvals.maxCoeff();
    
    // outputs
    lambda1normalized= eigvals(1);
    lambda2normalized= eigvals(2);
    projcloud= newcloud;
    projnormals= nproj;
}

float cloudAlignmentScoreDenseWithNormalsNormalized(const Cloud_t::Ptr &cloud1, const NormalCloud_t::Ptr &ncloud1,  const Cloud_t::Ptr &cloud2, const NormalCloud_t::Ptr &ncloud2, float relweight, float dnormalize) {
    
  vector<int> ix;
  vector<float> dd;
  Tree_t nntree;
  nntree.setInputCloud(cloud2);
  int N= cloud1->points.size();
  float accum=0;
  for(int i=0;i<N;i++) {
    nntree.nearestKSearch(cloud1->points[i], 1, ix, dd);
    
    accum += sqrt(dd[0]) / dnormalize; // normalize by provided length (usually extent of object)
    
    // also compare normals
    float dot= ncloud1->points[i].getNormalVector3fMap().dot(ncloud2->points[ix[0]].getNormalVector3fMap());
    accum += relweight*(1.0 - dot);
  }
  return accum/N;
}

// just return mean of distances from points in cloud1 to their nearest neighbors in cloud2 
float cloudAlignmentScoreDense(const Cloud_t::Ptr &cloud1, const Cloud_t::Ptr &cloud2) {
  
  vector<int> ix;
  vector<float> dd;
  Tree_t nntree;
  nntree.setInputCloud(cloud2);
  int N= cloud1->points.size();
  float accum=0;
  for(int i=0;i<N;i++) {
    nntree.nearestKSearch(cloud1->points[i], 1, ix, dd);
    accum += sqrt(dd[0]);
  }
  return accum/N;

}

struct RGBStruct{
  RGBStruct(float _rgb){ rgb = _rgb; }
  RGBStruct(){ s.r = 0; s.g = 0; s.b = 0; s.alpha = 0; }
  union{
    float rgb;
    struct{
      uint8_t b;
      uint8_t g;
      uint8_t r;
      uint8_t alpha;
    } s;
  };
};

float blend_colors( float rgb1f, float rgb2f )
{
  RGBStruct rgb1(rgb1f);
  RGBStruct rgb2(rgb2f);
  RGBStruct rgb_out;
  rgb_out.s.r = ((int)rgb1.s.r + (int)rgb2.s.r)/2;
  rgb_out.s.g = ((int)rgb1.s.g + (int)rgb2.s.g)/2;
  rgb_out.s.b = ((int)rgb1.s.b + (int)rgb2.s.b)/2;
  return rgb_out.rgb;
}
float blend_colors( float rgb1f, float rgb2f, float rgb3f )
{
  RGBStruct rgb1(rgb1f), rgb2(rgb2f), rgb3(rgb3f);
  RGBStruct rgb_out;
  rgb_out.s.r = ((int)rgb1.s.r + (int)rgb2.s.r + (int)rgb3.s.r)/3;
  rgb_out.s.g = ((int)rgb1.s.g + (int)rgb2.s.g + (int)rgb3.s.g)/3;
  rgb_out.s.b = ((int)rgb1.s.b + (int)rgb2.s.b + (int)rgb3.s.b)/3;
  return rgb_out.rgb;
}

Mesh_t::Ptr mesh_subdivide( const Mesh_t::ConstPtr &mesh, float thr) {
    
  // make a replicate for our new mesh
  Mesh_t::Ptr meshout(new Mesh_t);
  *meshout= *mesh;
  
  // unfold mesh, which is stored as ros message
  Cloud_t::Ptr pcl_cloud (new Cloud_t);
  pcl::fromROSMsg( mesh->cloud, *pcl_cloud );
  vector<pcl::Vertices> toadd;
  int processed= 0;
  for(size_t i=0;i<mesh->polygons.size();i++) {
    
    const pcl::Vertices& v= mesh->polygons[i];
    
    // Get the 3 points
    Point_t& p1= pcl_cloud->points[v.vertices[0]];
    Point_t& p2= pcl_cloud->points[v.vertices[1]];
    Point_t& p3= pcl_cloud->points[v.vertices[2]];
    
    // Convert to eigen points
    Vector3d pe1(p1.x,p1.y,p1.z);
    Vector3d pe2(p2.x,p2.y,p2.z);
    Vector3d pe3(p3.x,p3.y,p3.z);
    
    // Find length of maximal edge
    float l1= (pe1-pe2).norm();
    float l2= (pe2-pe3).norm();
    float l3= (pe3-pe1).norm();
    float maxedge= max(max(l1, l2), l3);
    if(maxedge < thr) continue; // below threshold, let's not subdivide
    
    // Find edge points of the polygon
    Vector3d c1= (pe1+pe2)/2;
    Vector3d c2= (pe2+pe3)/2;
    Vector3d c3= (pe1+pe3)/2;
    
    //Vector3d vcenter= (pe1 + pe2 + pe3)/3.0;
    //Point_t pcenter( vcenter[0], vcenter[1], vcenter[2] );
    
    // Subdivide this polygon
    
    pcl::Vertices pnew;
    pnew.vertices.resize(3);
    
    int dontadd= -1;
    size_t n1,n2;
    // find indices of largest 2 edges
    if(l3<=l1 && l3<=l2) { 
      Point_t pt1; pt1.x = c1[0]; pt1.y = c1[1]; pt1.z = c1[2];
      Point_t pt2; pt2.x = c2[0]; pt2.y = c2[1]; pt2.z = c2[2];
#ifdef USE_COLOR
      pt1.rgb = blend_colors(p1.rgb, p2.rgb);
      pt2.rgb = blend_colors(p2.rgb, p3.rgb);
#endif
      pcl_cloud->points.push_back(pt1);  
      pcl_cloud->points.push_back(pt2);
      n1= pcl_cloud->points.size() - 2;
      n2= pcl_cloud->points.size() - 1;
      
      pnew.vertices[0]= v.vertices[0];
      pnew.vertices[1]= v.vertices[2];
      pnew.vertices[2]= n2;
      
      dontadd= 2;
    }
    else if(l2<=l1 && l2<=l3) { 
      Point_t pt1; pt1.x = c3[0]; pt1.y = c3[1]; pt1.z = c3[2];
      Point_t pt2; pt2.x = c1[0]; pt2.y = c1[1]; pt2.z = c1[2];
#ifdef USE_COLOR
      pt1.rgb = blend_colors(p1.rgb, p3.rgb);
      pt2.rgb = blend_colors(p1.rgb, p2.rgb);
#endif
      pcl_cloud->points.push_back(pt1);
      pcl_cloud->points.push_back(pt2);        
      n1= pcl_cloud->points.size() - 2;
      n2= pcl_cloud->points.size() - 1;
      
      pnew.vertices[0]= v.vertices[2];
      pnew.vertices[1]= v.vertices[1];
      pnew.vertices[2]= n2;
      
      dontadd= 1;
    }
    else{ //if(l1<=l2 && l1<=l3) { 
      Point_t pt1; pt1.x = c2[0]; pt1.y = c2[1]; pt1.z = c2[2];
      Point_t pt2; pt2.x = c3[0]; pt2.y = c3[1]; pt2.z = c3[2];
#ifdef USE_COLOR
      pt1.rgb = blend_colors(p2.rgb, p3.rgb);
      pt2.rgb = blend_colors(p1.rgb, p3.rgb);
#endif
      pcl_cloud->points.push_back(pt1);
      pcl_cloud->points.push_back(pt2);        
      n1= pcl_cloud->points.size() - 2;
      n2= pcl_cloud->points.size() - 1;
      
      pnew.vertices[0]= v.vertices[1];
      pnew.vertices[1]= v.vertices[0];
      pnew.vertices[2]= n2;
      
      dontadd= 0;
    }
        
    //overwrite one polygon
    meshout->polygons[i]= pnew;
    
    // add other polygons to our toadd vector
    for(size_t j=0;j<3;j++) {
      if(j == dontadd) continue;
      pnew.vertices[0]= v.vertices[j];
      pnew.vertices[1]= n1;
      pnew.vertices[2]= n2;
      toadd.push_back(pnew);
    }
    
    processed++;
  } 
  //printf("Subdivided total of %d/%d meshes.\n", processed, mesh->polygons.size());
  
  //add all new polygons in toadd to mesh->polygons
  //meshout->polygons.insert(meshout->polygons.end(), toadd.begin(), toadd.end());
  for(size_t k=0;k<toadd.size();k++) meshout->polygons.push_back(toadd[k]);
  
  //printf("Old #polygons = %d, New #polygons = %d.\n", mesh->polygons.size(), meshout->polygons.size());
  
  // Also copy over the new cloud of points to new mesh
  pcl_cloud->width= pcl_cloud->points.size();
  pcl_cloud->height= 1;
  pcl::toROSMsg( *pcl_cloud, meshout->cloud );
  return meshout;
}

Mesh_t::Ptr mesh_subdivide_recursive( const Mesh_t::ConstPtr &mesh, float thr) {
  Mesh_t::Ptr mesh2(new Mesh_t);
  *mesh2= *mesh;
  while(true) {
    int num_before= mesh2->polygons.size();
    mesh2= mesh_subdivide(mesh2, thr);
    int num_after= mesh2->polygons.size(); 
    if(num_after==num_before) return mesh2;
  }
}

#endif

