#ifndef __OBJ_MEASURES_H__
#define __OBJ_MEASURES_H__

#include <pcl/surface/convex_hull.h>
#include <pcl/surface/impl/convex_hull.hpp>
#include "types.h"
#include "geom_utils.h"

// Objectness measures. They all assume that cloud is in its eigenbasis
// That is, it is longest along first dimension. 

// Compactness: ratio of surface area of the segment to surface area of its smallest bounding sphere
float scoreCompactness(Cloud_t::Ptr cloud) {
  
  // this silly implementation maximizes speed to readibility ratio
  float maxx=cloud->points[0].x;
  float minx=cloud->points[0].x;
  for(int i=1;i<cloud->points.size();i++) {
    float x=cloud->points[i].x;
    if(x>maxx)maxx=x;
    if(x<minx)minx=x;
  }
  return cloud->points.size() / pow(maxx-minx, 2);
}

// Symmetry: flip the cloud along all 3 principle axes and measure overlap
float scoreSymmetry(Cloud_t::Ptr cloud, NormalCloud_t::Ptr normals, float relweight) {
  
  Eigen::MatrixXf ptsmap = cloud->getMatrixXfMap();
  Eigen::VectorXf mins= ptsmap.rowwise().minCoeff(); 
  Eigen::VectorXf maxes= ptsmap.rowwise().maxCoeff(); 
  Eigen::VectorXf ranges= maxes - mins;
  Eigen::VectorXf w= ranges.head(3); // x,y,z are always the first 3 dimensions
  w /= w.sum();
  
  float score=0; 
  Cloud_t::Ptr dest (new Cloud_t);
  dest->points.resize(cloud->points.size());
  NormalCloud_t::Ptr normdest (new NormalCloud_t);
  normdest->points.resize(normals->points.size());
  for(int t=0;t<3;t++) {
  
    for(int i=0;i<cloud->points.size();i++){
      Point_t &d= dest->points[i];
      Point_t &c= cloud->points[i];
      Normal_t &dn= normdest->points[i];
      Normal_t &cn= normals->points[i];
      if(t==0) {d.x = -c.x; dn.normal_x = -cn.normal_x;} else {d.x = c.x; dn.normal_x = cn.normal_x;}
      if(t==1) {d.y = -c.y; dn.normal_y = -cn.normal_y;} else {d.y = c.y; dn.normal_y = cn.normal_y;}
      if(t==2) {d.z = -c.z; dn.normal_z = -cn.normal_z;} else {d.z = c.z; dn.normal_y = cn.normal_y;}
    }
    float overlap= cloudAlignmentScoreDenseWithNormalsNormalized(cloud, normals, dest, normdest, relweight, ranges(t))
                  +cloudAlignmentScoreDenseWithNormalsNormalized(dest, normdest, cloud, normals, relweight, ranges(t));
    score += w(t)*overlap;
  }
  return -score; // return the average symmetry
}


// global convexity: the convex hull should be a good approximation of the segment
float scoreGlobalConvexity(Cloud_t::Ptr cloud) {

  pcl::ConvexHull<Point_t> chull;
  chull.setInputCloud(cloud);
  chull.setDimension(3);
  vector<pcl::Vertices> polygons;
  Cloud_t::Ptr hull (new Cloud_t);
  chull.reconstruct(*hull, polygons);
  
  // make a full Mesh_t out of it
  Mesh_t::Ptr m (new Mesh_t);
  m->polygons= polygons;
  pcl::toROSMsg(*hull, m->cloud);
  
  // subdivide it 
  // 0.01 is not really a free parameter. The smaller the better and 1.5cm is reasonable.
  Mesh_t::Ptr msub= mesh_subdivide_recursive(m, 0.015); 
  Cloud_t::Ptr verts(new Cloud_t);
  pcl::fromROSMsg(msub->cloud, *verts);
  
  // calculate alignment from object to hull
  float overlap= cloudAlignmentScoreDense(cloud, verts);
  return -overlap;
}

// objects are usually composed of many locally convex regions
float scoreLocalConvexity(Cloud_t::Ptr cloud, NormalCloud_t::Ptr ncloud, float NNradius) {
  
  Tree_t nntree;
  nntree.setInputCloud(cloud);
  int N= cloud->points.size();
  vector<int> ix;
  vector<float> dd;
  float score= 0.0;
  for(int i=0;i<N;i++) {
    nntree.radiusSearch(i, NNradius, ix, dd);
    Normal_t &n1= ncloud->points[i];
    Point_t &p1= cloud->points[i];
    float dots= 0.0;
    int nc=0;
    for(int j=0;j<ix.size();j++) {
      if(ix[j]==i) continue;
      Normal_t &n2= ncloud->points[ix[j]];
      Point_t &p2= cloud->points[ix[j]];
      
      float dot= n1.normal_x*n2.normal_x + n1.normal_y*n2.normal_y + n1.normal_z*n2.normal_z;
      
      float a= p2.x-p1.x;
      float b= p2.y-p1.y;
      float c= p2.z-p1.z;
      float d= sqrt(a*a+b*b+c*c);
      a/=d;b/=d;c/=d;
      float dot2= n2.normal_x*a + n2.normal_y*b + n2.normal_z*c;
      
      if(dot2>0.0) nc++;
    }
    score += (1.0*nc)/ix.size();
  }
  return score/N; 
}

// Projects local points to tangent plane of every point and is high when
// there is a lot of mass in all directions around a point. i.e. 
// entropy is high of the angle distribution histogram of surrounding mass in 
// the local 2D coordinate system.
float scoreSmoothness(Cloud_t::Ptr cloud, NormalCloud_t::Ptr ncloud, float NNradius, int NUMBINS) {
  
  Tree_t nntree;
  nntree.setInputCloud(cloud);
  int N= cloud->points.size();
  vector<int> ix;
  vector<float> dd;
  float Hs=0.0;
  for(int i=0;i<N;i++) {
    
    // find a vector perpendicular to n1 by orthogonalizing
    Normal_t &n1= ncloud->points[i];
    Normal_t n2(randf(-1,1), randf(-1,1), randf(-1,1));
    float dot= n1.normal_x*n2.normal_x + n1.normal_y*n2.normal_y + n1.normal_z*n2.normal_z;
    n2.normal_x=n2.normal_x-dot*n1.normal_x;
    n2.normal_y=n2.normal_y-dot*n1.normal_y;
    n2.normal_z=n2.normal_z-dot*n1.normal_z;
    
    // normalize the orthogonal vector
    float ddd= sqrt(n2.normal_x*n2.normal_x + n2.normal_y*n2.normal_y + n2.normal_z*n2.normal_z);
    n2.normal_x/=ddd; n2.normal_y/=ddd; n2.normal_z/=ddd; 
    
    // find the second vector using cross product
    Normal_t n3;
    n3.normal_x= n1.normal_y*n2.normal_z - n1.normal_z*n2.normal_y;
    n3.normal_y= n1.normal_z*n2.normal_x - n1.normal_x*n2.normal_z;
    n3.normal_z= n1.normal_x*n2.normal_y - n1.normal_y*n2.normal_x;
    
    Point_t &p1= cloud->points[i];
    vector<float> bins(NUMBINS, 1.0);
    float binsum=NUMBINS;
    nntree.radiusSearch(i, NNradius, ix, dd);
    for(int j=0;j<ix.size();j++) {
      if(ix[j]==i) continue;
      Point_t &p2= cloud->points[ix[j]];
      float a= p2.x-p1.x;
      float b= p2.y-p1.y;
      float c= p2.z-p1.z;
      float d= sqrt(a*a+b*b+c*c);
      a/=d;b/=d;c/=d;
      
      float dot1= n2.normal_x*a + n2.normal_y*b + n2.normal_z*c;
      float dot2= n3.normal_x*a + n3.normal_y*b + n3.normal_z*c;
      float theta= (atan2(dot1, dot2) + M_PI)/2.0/M_PI; // angle in range 0->1
      
      int binid= (int)((theta-0.001)*NUMBINS); // subbing 0.001 is an ugly hack but that's ok. It prevents binid=NUMBINS
      bins[binid] += 1.0;
      binsum += 1.0;
    }
    
    float H=0.0;
    for(int j=0;j<NUMBINS;j++) bins[j]/=binsum;
    for(int j=0;j<NUMBINS;j++) H -= bins[j]*log(bins[j]);
    Hs += H;
  }
  return Hs/N; // high entropy = good.
}

#endif
