
#include <pcl/io/vtk_lib_io.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "types.h"
#include "utils.h"
#include "io_utils.h"
#include "vis_wrapper.h"
#include "geom_utils.h"
#include "segment-graph.h" // felzenswalb segmenter
#include "measures.h"
#include "eigen_extensions.h"

using namespace std;

// Take a scene id and segment the scene to its object hypotheses
//  outputs .ply files for object hypotheses and measures.eig.txt file that contains the objectness measures
int main(int argc, char** argv) {
  srand(1);

  assert(argc > 1);
  int sceneid= atoi(argv[1]);
  string plyfile= getScenePath(sceneid);
  if(!file_exists(plyfile)) {
    cout << "EEEeee, file " << plyfile << " does not exist. Quitting." << endl;
    return(0);
  }
  
  int dosave= 1;
  if(argc > 2) dosave= atoi(argv[2]);
  
  int DISPLAY= 1;
  if(argc > 3) DISPLAY= atoi(argv[3]);
  
  // MAGIC NUMBERS! yay
  //----------------------------------------------------------------------------
  
  // segmentation parameters
  vector<float> KTHRS;
  KTHRS.push_back(0.7);
  KTHRS.push_back(0.8);
  KTHRS.push_back(0.9);
  KTHRS.push_back(1.0);
  KTHRS.push_back(1.1);
  KTHRS.push_back(1.2);
  KTHRS.push_back(1.5);
  KTHRS.push_back(1.7);
  KTHRS.push_back(2.0);
  KTHRS.push_back(3.0);
  int MINSEGSIZE= 500; // minimum segment size, or it is merged to neighboring clusters during segmentation
  
  // hard objectness parameters
  int MINPTS= 750; // max and min number of points. These two are mostly just for efficiency
  int MAXPTS= 30000;
  
  float THINTHR= 0.05;
  float FLATTHR= 0.001;
  
  float MINLENGTH= 0.02; // minimum extent of a segment along minimal direction (in meters)
  float MAXLENGTH= 0.85; // maximum extent of a segment along maximal direction (in meters)
  
  // measure parameters
  float SYMMETRY_MEASURE_CLOUD_NORMALS_TRADEOFF= 0.2; // scaling factor for difference in normals wrt. difference in position for points,
                                      // when computing difference between two point clouds. 
                                      // 0 => Only look at difference between point and its closest point
                                      // Higher value => matching normals are more important than point distances
  float LOCAL_CONVX_MEASURE_NNRADIUS = 0.0075; // used in Local Convexity measure for local neighborhood
  float SMOOTHNESS_MEASURE_NNRADIUS = 0.01; // used in Smoothness measure for local neighborhood finding
  float SMOOTHNESS_MEASURE_NUMBINS = 8; // Number of bins in histogram. We found 8 to work best quite consistently.
  
  // non-max suppression parameters       
  float IOUTHR= 0.8; // intersection over union threshold, during non-max suppression
  //----------------------------------------------------------------------------
  
  MAKECLOCK;
  VisWrapper v;
  v.vis_.setBackgroundColor(0.5, 0.5, 0.5);
  v.vis_.setCameraPose(0, 0, 0, 
                       0.3, 0.3, 0.3, 
                       0, -1, 0);
       
  // Load the geometry from .ply and extract the mesh, cloud, normals
  TIC;
  Mesh_t::Ptr sceneMesh(new Mesh_t);
  Cloud_t::Ptr sceneCloud(new Cloud_t);
  NormalCloud_t::Ptr sceneNormals (new NormalCloud_t);
  cout << "Loading file " << plyfile << endl;
  pcl::io::loadPolygonFile(plyfile, *sceneMesh);
  mesh_to_points_normals(sceneMesh, sceneCloud, sceneNormals);
  TOC("loaded geometry");
  
  if(DISPLAY==1) {
    printf("Displaying the scene cloud. Press 'q' to continue\n");
    v.showCloud(sceneCloud);
    while(v.waitKey() != 'q') continue;
  }
  
  // Construct the edge graph for segmentation
  printf("Constructing edge graph based on mesh connectivity...\n");
  TIC;
  int Npoly = sceneMesh->polygons.size();
  int N = sceneCloud->points.size();
  vector<pair<int,int> >es; getEdges(sceneMesh, es);
  int Ne = es.size();
  edge *edges = new edge[Ne];
  for(int i=0;i<Ne;i++) {
    
    int a= es[i].first;
    int b= es[i].second;
    
    edges[i].a = a;
    edges[i].b = b;
    
    Normal_t &n1= sceneNormals->points[a];
    Normal_t &n2= sceneNormals->points[b];
    Point_t &p1= sceneCloud->points[a];
    Point_t &p2= sceneCloud->points[b];
    
    float dx= p2.x-p1.x;
    float dy= p2.y-p1.y;
    float dz= p2.z-p1.z;
    float dd= sqrt(dx*dx+dy*dy+dz*dz); dx/=dd;dy/=dd;dz/=dd;
    float dot= n1.normal_x*n2.normal_x + n1.normal_y*n2.normal_y + n1.normal_z*n2.normal_z;
    float dot2= n2.normal_x*dx + n2.normal_y*dy + n2.normal_z*dz;
    float ww= 1.0 - dot;
    if(dot2>0) ww= ww*ww; // make it much less of a problem if convex regions have normal difference
    edges[i].w= ww;
  }
  TOC("made graph");
  
  // Go over all thresholds, segment, accumulate all segments into a bag
  printf("Segmenting scene with over many granularity thresholds...\n");
  vector<Cloud_t::Ptr> allsegments; // will store all clouds across all thresholds
  vector<NormalCloud_t::Ptr> allsegmentNormals; // the associated normals
  vector<set<int> > allsegmentpts; // the point indeces. Will be used to compute union over intersection overlap
  for(int kk=0;kk<KTHRS.size();kk++) {
    
    // Segment!
    float kthr= KTHRS[kk];
    TIC; universe *u = segment_graph(N, Ne, edges, kthr); TOC("segmented");
    
    // post-process segmentation by joining small segments
    for(int j=0;j<Ne;j++) {
      int a = u->find(edges[j].a);
      int b = u->find(edges[j].b);
      if ((a != b) && ((u->size(a) < MINSEGSIZE) || (u->size(b) < MINSEGSIZE))) {
        u->join(a, b);
      }
    }
    
    // display colored point cloud
    if(DISPLAY == 1) {
      map<int, rgb> colmap;
      CloudColor_t::Ptr cres (new CloudColor_t);
      PointColor_t pt;
      rgb blah;
      for(int q=0;q<N;q++) {
        
        int comp = u->find(q);
        if(colmap.count(comp)>0) blah= colmap[comp];
        else {
          // add new color
          blah.r= (rand() % 205)+50;
          blah.g= (rand() % 205)+50;
          blah.b= (rand() % 205)+50;
          colmap[comp]= blah;
        }
        
        pt.x= sceneCloud->points[q].x;
        pt.y= sceneCloud->points[q].y;
        pt.z= sceneCloud->points[q].z;
        pt.r= blah.r;
        pt.g= blah.g;
        pt.b= blah.b;
        
        cres->push_back(pt);
      }
      v.showCloud(cres);
      printf("Displaying segmentation at threshold k=%f. Press 'q' to continue.\n", kthr);
      while(v.waitKey() != 'q') continue;
    }
    
    // place all segments into their own cloud
    TIC;
    vector<Cloud_t::Ptr> segments;
    vector<NormalCloud_t::Ptr> segmentNormals;
    vector<set<int> > segmentpts; // for each segment we'll keep set of points assigned to it. For duplicate detection
    map<int, int> compToIx;
    for(int q=0;q<N;q++) {
      
      int comp = u->find(q);
      int sz= u->size(comp);
      
      // done for efficiency here
      if(sz < MINPTS) continue; // segment way too small
      if(sz > MAXPTS) continue; // segment way too big
      
      if(compToIx.count(comp) > 0) {
      
        // add it on top of the (already created) cloud for this segment
        int mapix= compToIx[comp];
        Cloud_t::Ptr &m = segments[mapix];
        m->push_back(sceneCloud->points[q]);
        m->height++;
        
        NormalCloud_t::Ptr &nm = segmentNormals[mapix];
        nm->push_back(sceneNormals->points[q]);
        nm->height++;

        segmentpts[compToIx[comp]].insert(q);
        
      } else {
      
        // create new cloud for this segment
        Cloud_t::Ptr m = Cloud_t::Ptr(new Cloud_t);
        m->height = 1;
        m->width = 1;
        m->push_back(sceneCloud->points[q]);
        segments.push_back(m);
        
        // create new normals cloud for this segment
        NormalCloud_t::Ptr nm = NormalCloud_t::Ptr(new NormalCloud_t);
        nm->height = 1;
        nm->width = 1;
        nm->push_back(sceneNormals->points[q]);
        segmentNormals.push_back(nm);
        
        // create new set of indeces for this segment
        set<int> s; s.insert(q);
        segmentpts.push_back(s);
        compToIx[comp]= segments.size() - 1;
      }
    }
    
    int Ncomp = segments.size();
    printf("Found %d components using k=%f!\n", Ncomp, kthr);
    
    // append into our bags that keep track of segments across all thresholds
    for(int q=0;q<segments.size();q++) allsegments.push_back(segments[q]);
    for(int q=0;q<segmentNormals.size();q++) allsegmentNormals.push_back(segmentNormals[q]);
    for(int q=0;q<segmentpts.size();q++) allsegmentpts.push_back(segmentpts[q]);
    
  } // end loop over segmentation thresholds
  TOC("made segments");
  delete [] edges; // clean up this space, not needed anymore
  
  // Project all clouds to their eigenbasis
  printf("Projecting all segments to their eigenbasis...\n");
  int NS= allsegments.size();
  vector<int> keep(NS, 1); // keep is an indicator vector of what segments we will keep at very end
  vector<Cloud_t::Ptr> cloudBuffer; // for visualization only, keeps track of segments we remove
  for(int i=0;i<NS;i++) {
    float lambda1n, lambda2n;
    projectToEigenbasis(allsegments[i], allsegmentNormals[i],
                        allsegments[i], allsegmentNormals[i],
                        lambda1n, lambda2n);
    
    // while we're at it, lets eliminate segments that are too big or too small (relatively or absolutely)
    if(lambda1n < THINTHR) {
      keep[i]=0; // Object way too thin
      printf("Eliminating segment %d because eigvals(1)/eigvals(0)=%f, so it is waaay too thin.\n", i, lambda1n);
      cloudBuffer.push_back(allsegments[i]);
      continue;
    }
    if(lambda2n < FLATTHR) {
      keep[i]=0; // Object way too flat
      printf("Eliminating segment %d because eigvals(2)/eigvals(0)=%f, so it is waaay too flat.\n", i, lambda2n);
      cloudBuffer.push_back(allsegments[i]);
      continue;
    }
    
    // the cloud is now in eigenbasis so ranges(0) is longest length and ranges(2) is shortest length
    Eigen::MatrixXf ptsmap = allsegments[i]->getMatrixXfMap();
    Eigen::VectorXf mins= ptsmap.rowwise().minCoeff(); 
    Eigen::VectorXf maxes= ptsmap.rowwise().maxCoeff(); 
    Eigen::VectorXf ranges= maxes - mins;
    float vx= ranges(0); float vy= ranges(1); float vz= ranges(2);
    if(vx < MINLENGTH || vy < MINLENGTH || vz < MINLENGTH || vx > MAXLENGTH || vy > MAXLENGTH || vz > MAXLENGTH) {
      keep[i]= 0; 
      printf("Eliminating cloud %d because its too small/big in some direction %f %f %f\n", i, vx, vy, vz);
      cloudBuffer.push_back(allsegments[i]);
      continue;
    }
  }
  
  if(DISPLAY == 1) {
    printf("Displaying segments that were rejected due to dimensions. Press 'q' to continue.\n");
    showCloudsGrid(v, cloudBuffer);
    while(v.waitKey()!='q') continue;
  }
  
  // Compute objectness measures on all segments
  printf("Computing shape objectness measures on all segments...\n");
  int nummeasures= 5;
  Eigen::MatrixXf measures(NS, nummeasures); measures.setZero();
  Eigen::MatrixXf mrows(NS, nummeasures); mrows.setZero();
  Eigen::VectorXf times(nummeasures); times.setZero();
  int np=0;
  for(int i=0;i<NS;i++) {
    if(i%5==0) printf("Processing segment %d/%d\n", i, NS);
    if(keep[i]==0) continue;
    double s;
    s=pcl::getTime(); measures(i,0)= scoreCompactness(allsegments[i]); times(0)+=pcl::getTime()-s;
    s=pcl::getTime(); measures(i,1)= scoreSymmetry(allsegments[i], allsegmentNormals[i], SYMMETRY_MEASURE_CLOUD_NORMALS_TRADEOFF); times(1)+=pcl::getTime()-s;
    s=pcl::getTime(); measures(i,2)= scoreGlobalConvexity(allsegments[i]); times(2)+=pcl::getTime()-s;
    s=pcl::getTime(); measures(i,3)= scoreLocalConvexity(allsegments[i], allsegmentNormals[i], LOCAL_CONVX_MEASURE_NNRADIUS); times(3)+=pcl::getTime()-s;
    s=pcl::getTime(); measures(i,4)= scoreSmoothness(allsegments[i], allsegmentNormals[i], SMOOTHNESS_MEASURE_NNRADIUS, SMOOTHNESS_MEASURE_NUMBINS); times(4)+=pcl::getTime()-s;
    
    mrows.row(np)= measures.row(i);
    np++;  
  }
  times /= 0.001*np;
  printf("Average compute times (in milliseconds) per segment:\n");
  printf("compactness: %.1fms\n", times(0));
  printf("symmetry: %.1fms\n", times(1));
  printf("global convexity: %.1fms\n", times(2));
  printf("local convexity: %.1fms\n", times(3));
  printf("smoothness: %.1fms\n", times(4));
  
  // Compute mean and standard deviation along each column
  Eigen::MatrixXf mm= mrows.block(0,0,np,nummeasures);
  Eigen::VectorXf mmean= mm.colwise().sum() / mm.rows();
  mm.rowwise() -= mmean.transpose();
  Eigen::MatrixXf mstd= (mm.colwise().squaredNorm() / mm.rows()).cwiseSqrt();
  
  // Normalize the data
  Eigen::MatrixXf measuresNormalized( measures );
  measuresNormalized.rowwise() -= mmean.transpose();
  measuresNormalized = measuresNormalized.cwiseQuotient(mstd.replicate(measuresNormalized.rows(), 1));
  Eigen::VectorXf objectness= measuresNormalized.rowwise().sum() / measuresNormalized.cols(); // objectness is just the average
  
  // Carry out non-maximum suppression
  printf("Eliminating duplicate segments using non-max suppression based on objectness...\n");
  TIC;
  int ntot=0;
  int nevaled=0;
  for(int i=0;i<allsegmentpts.size();i++) {
    set<int> &s1= allsegmentpts[i];
    for(int j=0;j<allsegmentpts.size();j++) {
      if(i==j) continue;
      if(keep[i]==0) continue;
      if(keep[j]==0) continue;
      set<int> &s2= allsegmentpts[j];
      
      float s1s= s1.size();
      float s2s= s2.size();
      float bestCaseIOU= min(s1s, s2s) / max(s1s, s2s);
      ntot++;
      //printf("%d, %d: bestCaseIOU = %f\n", i, j, bestCaseIOU);
      
      if(bestCaseIOU > IOUTHR) {
        nevaled++;
        // doublecheck
        set<int> intersect, setunion;
        set_intersection( s1.begin(), s1.end(), s2.begin(), s2.end(), inserter(intersect, intersect.begin()));
        set_union( s1.begin(), s1.end(), s2.begin(), s2.end(), inserter(setunion, setunion.begin()));
        
        float IOU = (1.0*intersect.size()) / setunion.size();
        if( IOU > IOUTHR ) {
        
          // too much overlap, retain the one with higher objectness
          // also give advantage to bigger segments
          if(objectness(i) > objectness(j)) {
            keep[j] = 0;
            printf("%d is suppressing %d because %f > %f\n", i, j, objectness(i), objectness(j));
          } else {
            keep[i] = 0;
            printf("%d is suppressing %d because %f <= %f\n", j, i, objectness(i), objectness(j));
          }
        }
      }
    }
  }
  printf("Total number of duplicate checks: %d, but only %d were expensively considered.\n", ntot, nevaled);
  TOC("non-max suppression computed");
  int NSkeep=0; for(int i=0;i<NS;i++) { if(keep[i]==1) NSkeep++; } // count up how many segments we're keeping in total
  
  // show final objects found in this scene highlighted in the scene
  if(DISPLAY==1) {
    printf("Displaying top segments found in the scene. Use 'w'/'s' to show more or less, and use 'q' to continue.\n");
    
    // sort all segments by objectness
    vector<fipair> objv(NS);
    for(int i=0;i<NS;i++) objv[i]= fipair(objectness(i), i);
    sort(objv.begin(), objv.end(), fiComparatorDescend);

    int numwant= 0;
    Cloud_t::Ptr sc(new Cloud_t(*sceneCloud));
    v.holdOff();
    while(true) {
      
      // dim the original scene cloud
      float dimratio= 0.2;
      for(int i=0;i<sc->points.size();i++) {
        sc->points[i].r = floor(dimratio * sceneCloud->points[i].r);
        sc->points[i].g = floor(dimratio * sceneCloud->points[i].g);
        sc->points[i].b = floor(dimratio * sceneCloud->points[i].b);
      }
      
      // iterate over segments and make their points visible in original
      int numshowed=0;
      for(int iii=0;iii<allsegmentpts.size();iii++) {
        int ii= objv[iii].second;
        if(keep[ii]==0) continue;
        
        for(set<int>::const_iterator it= allsegmentpts[ii].begin(); it != allsegmentpts[ii].end(); it++) {
          int i= *it;
          sc->points[i].r = sceneCloud->points[i].r;
          sc->points[i].g = sceneCloud->points[i].g;
          sc->points[i].b = sceneCloud->points[i].b;
        }   
        numshowed++;
        if(numshowed>=numwant) break;
      }
    
      v.showCloud(sc);
      bool escape= false;
      while(true) {
      char ch=v.waitKey();
        if(ch=='q') {escape= true; break;}
        if(ch=='w') {numwant++; break;}
        if(ch=='s') {numwant--; break;}
      }
      if(numwant<0)numwant=0;
      if(escape) break;
    }
  }
  
  if(DISPLAY==1) {
    printf("Displaying grid of segments in order of decreasing objectness.\n");
    
    int om=0;
    bool redraw= true;
    while(true) {
      
      if(redraw) {
      
        // sort by objectness
        vector<fipair> objv(NS);
        printf("0: Average, 1: Compactness, 2: Symmetry, 3: Global Convexity, 4: Local Convexity, 5: Smoothness\n");
        printf("currently sorted by %d\n", om % 6);
        if(om % 6 == 0) { for(int i=0;i<NS;i++) objv[i]= fipair( objectness(i), i ); }
        if(om % 6 == 1) { for(int i=0;i<NS;i++) objv[i]= fipair( measuresNormalized(i,0), i ); }
        if(om % 6 == 2) { for(int i=0;i<NS;i++) objv[i]= fipair( measuresNormalized(i,1), i ); }
        if(om % 6 == 3) { for(int i=0;i<NS;i++) objv[i]= fipair( measuresNormalized(i,2), i ); }
        if(om % 6 == 4) { for(int i=0;i<NS;i++) objv[i]= fipair( measuresNormalized(i,3), i ); }
        if(om % 6 == 5) { for(int i=0;i<NS;i++) objv[i]= fipair( measuresNormalized(i,4), i ); }
        sort(objv.begin(), objv.end(), fiComparatorDescend);

        vector<Cloud_t::Ptr> toshowc;
        vector<NormalCloud_t::Ptr> toshown;
        for(int ii=0;ii<NS;ii++) {
          int i= objv[ii].second;
          if(keep[i]==0) continue;
          toshowc.push_back(allsegments[i]);
          toshown.push_back(allsegmentNormals[i]);
          printf("Showing cloud %d with value %f\n", ii, objv[ii].first);
        }
        showCloudsGridOriented(v, toshowc, toshown);
        
      }
      redraw= false;
      char ch= v.waitKey();
      if(ch=='q') break;
      if(ch=='w') { om++; redraw= true; }// cycle through objectness measures
      if(ch=='s') { om--; redraw= true; }
    }
  }
  
  // Now we have to do some annoying book-keeping. We have clouds, indeces, and normals of all segments
  // and keep[] indicator vector is telling us which ones we want to keep. We need to use the original
  // mesh to pluck out these segments into their own little mesh segments so that we can save them.
  printf("Creating meshes for all segments...\n");
  
  // first create a lookup map of vertex -> indices of all polygons each vertex is involved in
  TIC;
  map<int, vector<int> > vmap;
  for(int j=0; j < sceneMesh->polygons.size(); j++) { 
    const pcl::Vertices& v= sceneMesh->polygons[j];
    int v1 = v.vertices[0];
    int v2 = v.vertices[1];
    int v3 = v.vertices[2];
    vmap[v1].push_back(j);
    vmap[v2].push_back(j);
    vmap[v3].push_back(j);
  }
  
  // now stitch all mesh segments together
  vector<Mesh_t::Ptr> meshes(NS);
  for(int i=0;i<allsegments.size();i++) {
    if(keep[i]==0) continue;
      
    // we are keeping this segment, so convert it to mesh
    Mesh_t::Ptr newmesh (new Mesh_t);
    set<int> &ss= allsegmentpts[i];
    Cloud_t::Ptr currpts= allsegments[i];
    
    // pass 1: for all points in this segment, copy over the vertices from sceneCloud
    map<int,int> ixmap;
    int nc=0;
    for(set<int>::const_iterator it= ss.begin(); it != ss.end(); it++) {
      ixmap[*it]= nc;
      nc++;
    }
    
    // pass 2: count up for polygons, we only want to copy over those st all points 
    // are in our set
    vector<int> counts(sceneMesh->polygons.size(), 0);
    for(set<int>::const_iterator it= ss.begin(); it != ss.end(); it++) {
      vector<int> &vi= vmap[*it]; // get indeces of all polygons this vertex is in
      for(int j=0;j<vi.size();j++) { counts[vi[j]]++; }
    }
    
    // pass 3: copy over the associated polygons
    for(set<int>::const_iterator it= ss.begin(); it != ss.end(); it++) {
      vector<int> &vi= vmap[*it]; // get indeces of all polygons this vertex is in
      for(int j=0;j<vi.size();j++) {
        if(counts[vi[j]] < 3) continue;
        
        pcl::Vertices& v= sceneMesh->polygons[vi[j]];
        pcl::Vertices pnew;
        pnew.vertices.resize(3);
        pnew.vertices[0]= ixmap[v.vertices[0]];
        pnew.vertices[1]= ixmap[v.vertices[1]];
        pnew.vertices[2]= ixmap[v.vertices[2]];
        
        newmesh->polygons.push_back(pnew);
        counts[vi[j]]= 0; // set to 0 so that other vertex doesn't readd this polygon
      }
    }
    pcl::toROSMsg(*(allsegments[i]), newmesh->cloud);
    meshes[i] = newmesh;
  }
  TOC("stitching done");
  
  if(dosave == 1) {
  
    // save measures to savepath/measures and segments to savepath/segments
    Eigen::MatrixXf savemat(NSkeep, nummeasures);
    int n=0;
    for(int i=0;i<NS;i++) {
      if(keep[i]==0) continue;
      
      savemat.row(n)= measures.row(i);
      
      string segplypath= getSegmentPath(sceneid, n);
      pcl::io::savePolygonFile(segplypath, *(meshes[i]));
      cout << "saving segment " << n << " to " << segplypath << endl;
     
      n++;
    }
     
    string measurepath= getMeasurePath(sceneid);
    eigen_extensions::saveASCII(savemat, measurepath);
    cout << "saving measures to " << measurepath << endl;
  }
  
  printf("done!\n");
  return(0);
}
